package org.bdware.sc;

import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import com.google.gson.reflect.TypeToken;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractResult.Status;
import org.bdware.sc.bean.Contract;
import org.bdware.sc.bean.ContractExecType;
import org.bdware.sc.bean.FunctionDesp;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.conn.SocketGet;
import org.bdware.sc.db.CMTables;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.encrypt.HardwareInfo;
import org.bdware.sc.encrypt.HardwareInfo.OSType;
import org.bdware.sc.event.REvent.REventSemantics;
import org.bdware.sc.node.AnnotationNode;
import org.bdware.sc.node.YjsType;
import org.bdware.sc.util.JsonUtil;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.util.*;

public class ContractClient {
    private static final Logger LOGGER = LogManager.getLogger(ContractClient.class);
    public static String cmi = "";
    public ContractMeta contractMeta;
    public boolean isDebug;
    transient SocketGet get;
    int port;
    String pid;
    transient ContractPrinter outputTracer;
    transient ContractPrinter errorTracer;
    Date timeTravel; // 当前所处时间
    transient Process process;
    long times = 0L;
    long traffic = 0L;
    long memory;
    //    boolean withInsnLimit; // 是否需要计算gas
    ContractStatus contractStatus;
    boolean isRunning;

    //     public boolean changeDumpPeriod(String period) {
    //        System.out.println("[ContractClient] changeDumpPeriod " + period);
    //        String res = get.syncGet("", "changeDumpPeriod",period);
    //        if(res.equals("success"))
    //            return true;
    //        return false;
    //    }
    long lastUpdate = 0;

    ContractClient(Contract c) {
        contractMeta = new ContractMeta();
        contractMeta.contract = c;
        contractMeta.id = contractMeta.contract.getID();
        // contractMeta will be rewrite later
        outputTracer = new ContractPrinter();
        errorTracer = new ContractPrinter();
        contractStatus = ContractStatus.Ready;
        memory = 0;
        timeTravel = new Date();
        isRunning = false;
    }

    // Once the reconnect is called, all cp in the same node will reconnect to
    // current ContractManager
    // We only consider the.
    public ContractClient(int startPort) {
        // LOGGER.info("ContractClient  构造器 : ");
        contractMeta = new ContractMeta();
        outputTracer = new ContractPrinter();
        errorTracer = new ContractPrinter();
        contractStatus = ContractStatus.Ready;
        memory = 0;
        timeTravel = new Date();
        isRunning = false;
        port = startPort;
        // LOGGER.info("ContractClient----构造器----- 端口 " + startPort);

        get = new SocketGet("127.0.0.1", startPort);
        // LOGGER.info("ContractClient----构造器----- position---2");

        String cpCMI = get.syncGet("", "isContractProcess", ""); // CMI

        if (!cmi.equals(cpCMI)) {
            LOGGER.warn("contract listened to port " + startPort + " cpCMI = " + cpCMI);
            // LOGGER.info("ContractClient----构造器----- position---4");
            return;
        }
        pid = get.syncGet("", "getPID", "");
        // LOGGER.info("ContractClient----构造器----- position---5");
        initProps();

        // LOGGER.info("ContractClient----构造器----- position---6");
        // LOGGER.info("ContractClient----构造器----- position---7");
    }

    static boolean isBundlePath(String scriptStr) {
        return scriptStr.endsWith(".zip") || scriptStr.endsWith(".ypk");
    }

    public static String getPid(Process process) {
        long pid = -1;
        Field field;
        try {
            Class<?> clazz = Class.forName("java.lang.UNIXProcess");
            field = clazz.getDeclaredField("pid");
            field.setAccessible(true);
            pid = (Integer) field.get(process);
        } catch (Throwable e) {
            e.printStackTrace();
        }
        return String.valueOf(pid);
    }

    public boolean isProcessAlive() {
        return get != null && get.isAlive();
    }

    public boolean killProcess() {
        process.destroy();
        return true;
    }

    private void initProps() {
        // LOGGER.info("initProps  ----  position-----1");
        contractMeta.name = get.syncGet("", "getContractName", "");
        // LOGGER.info("initProps  ----  position-----2");
        String strEvent = get.syncGet("", "getDeclaredEvents", "");
        LOGGER.debug("event: " + strEvent);
        contractMeta.declaredEvents =
                JsonUtil.fromJson(
                        strEvent, new TypeToken<Map<String, REventSemantics>>() {
                        }.getType());
        // LOGGER.info("initProps  ----  position-----3");
        contractMeta.dependentContracts = JsonUtil.fromJson(
                get.syncGet("", "getDependentContracts", ""),
                new TypeToken<Set<String>>() {
                }.getType());
        // LOGGER.info("initProps  ----  position-----4");
        contractMeta.exportedFunctions = new HashMap<>();
        List<FunctionDesp> exportedFunctions =
                JsonUtil.fromJson(
                        get.syncGet("", "getExportedFunctions", ""),
                        new TypeToken<List<FunctionDesp>>() {
                        }.getType());
        contractMeta.logDetail = new HashMap<>();
        for (FunctionDesp func : exportedFunctions) {
            contractMeta.exportedFunctions.put(func.functionName, func);
            StringBuilder str = new StringBuilder();
            for (AnnotationNode anno : func.annotations) {
                if (anno.getType().equals("LogType")) {
                    for (String logType : anno.getArgs()) {
                        str.append(logType.replaceAll("\"", "")).append(";");
                    }
                }
                if (anno.getType().equals("LogLocation")) {
                    for (String logLoc : anno.getArgs()) {
                        if (logLoc.equals("\"dataware\"")
                                || logLoc.equals("\"bdledger\"")
                                || logLoc.equals("\"bdledger:\"")) {
                            str.append("bdcontract;");
                        } else if (logLoc.startsWith("\"bdledger:") && logLoc.length() > 11) {
                            String tmp = logLoc.substring(1);
                            String tmp2 = tmp.substring(0, tmp.length() - 1);
                            str.append(tmp2).append(";");
                        }
                    }
                }
            }
            contractMeta.logDetail.put(func.functionName, str.toString());
        }
        // LOGGER.info("initProps  ----  position-----5");
        try {

            String anno = get.syncGet("", "getAnnotations", "");
            contractMeta.annotations =
                    JsonUtil.fromJson(anno, new TypeToken<List<AnnotationNode>>() {
                    }.getType());

        } catch (Exception e) {
            // supoort contract process before version 0.70
            contractMeta.annotations = new ArrayList<>();
        }
        // LOGGER.info("initProps  ----  position-----6");
        contractMeta.sigRequired = Boolean.parseBoolean(get.syncGet("", "isSigRequired", ""));
        // LOGGER.info("initProps  ----  position-----7");
        contractMeta.thisPermission = get.syncGet("", "showPermission", "");
        // LOGGER.info("initProps  ----  position-----8");
        isRunning = true;
        contractMeta.isDebug = Boolean.parseBoolean(get.syncGet("", "getDebug", ""));
        // LOGGER.info("initProps  ----  position-----9");
        get.syncGet("", "registerMangerPort", ContractManager.cPort.getCMPort() + "");
        contractMeta.contract =
                JsonUtil.fromJson(get.syncGet("", "getContract", ""), Contract.class);
        contractMeta.id = contractMeta.contract.getID();
        get.setOfflineExceptionHandler(
                new SocketGet.OfflineHandler() {
                    @Override
                    public void onException(SocketGet socketGet, Exception e) {
                        if (e.getMessage().contains("Connection refused")) {
                            contractMeta.status = ContractStatusEnum.HANGED;
                            ContractManager.instance.statusRecorder.resumeContractProcess(
                                    contractMeta.id);
                        }
                    }
                });
        loadTimesAndTraffic();
        // LOGGER.info("initProps  ----  position-----10");
        // LOGGER.debug("======= registerPort:" + ret + "-->" + ContractManager.startPort);
    }

    public String startProcess(PrintStream ps) {
        isRunning = false;
        port = -1;

        String darg = "-Djava.library.path=";
        String classpath;
        File jniPath;

        if (ContractManager.yjsPath == null) {
            jniPath = new File("./jni/");
            classpath = System.getProperty("java.class.path");
        } else {
            classpath = ContractManager.yjsPath;
            jniPath = new File(classpath).getParentFile();
        }

        String osJni = ((HardwareInfo.type == OSType.linux) ? "/jni/linux" : "/jni/mac");
        darg += jniPath.getAbsolutePath() + osJni;
        if (!new File(classpath).exists()) {
            ContractResult r =
                    new ContractResult(
                            Status.Exception, new JsonPrimitive("incorrect path: yjs.jar"));
            return JsonUtil.toJson(r);
        }
        if (!new File(jniPath, "libs").exists()) {
            ContractResult r =
                    new ContractResult(
                            Status.Exception,
                            new JsonPrimitive("incorrect path: yjs.jar, missing libs"));
            return JsonUtil.toJson(r);
        }
        //        ProcessBuilder builder =
        //                new ProcessBuilder(
        //                        "java",
        //                        "-Dfile.encoding=UTF-8",
        //                        darg,
        //                        "-cp",
        //                        jniPath.getAbsolutePath() + "/libs/*:" + classpath,
        //                        "org.bdware.sc.ContractProcess",
        //                        "-port=" + cPort.getPort());
        int startPort = ContractManager.cPort.getPortAndInc();
        ProcessBuilder builder =
                new ProcessBuilder(
                        "java",
                        "-Dfile.encoding=UTF-8",
                        darg,
                        "-jar",
                        classpath,
                        "-port=" + startPort,
                        "-cmi=" + cmi, // cmi 区分不同CM的cp
                        (isDebug ? "-debug" : ""));
        File directory = new File("");
        LOGGER.debug("[CMD] path: " + directory.getAbsolutePath());
        LOGGER.debug(StringUtils.join(builder.command(), " "));

        Map<String, String> map = builder.environment();
        map.put("java.library.path", jniPath.getAbsolutePath() + osJni);
        builder.directory(new File("./"));
        LOGGER.debug("start process:");

        try {
            process = builder.start();

            this.pid = getPid(process);
            LOGGER.info("[CP PPID  ] " + pid);
            PrintStream printStream = new PrintStream(process.getOutputStream());
            printStream.println("CP PID:" + pid);
            printStream.close();
            Scanner sc = new Scanner(process.getInputStream());
            String status = null;
            while (sc.hasNext()) {
                status = sc.nextLine();
                LOGGER.info("[CP] " + status);
                if (status.contains("mainPort")) {
                    try {
                        // Set contractPort to max(mainPort, contractPort)
                        int portIndex = status.indexOf("mainPort");
                        int port =
                                Integer.parseInt(
                                        status.substring(portIndex + 9, portIndex + 14)
                                                .replaceAll("\\s+", ""));
                        if (port != startPort) {
                            ContractManager.cPort.reSetPort(port + 1);
                        }
                    } catch (Exception ignored) {
                    }
                    break;
                }
            }
            if (status != null) {
                status = status.replaceAll(".*mainPort ", "");
            }
            assert status != null;
            port = Integer.parseInt(status.split(" ")[0]);
            LOGGER.debug("port=" + port);
            ContractManager.cPort.updateDb(port, true);
            get = new SocketGet("127.0.0.1", port);
            get.syncGet("", "setDBInfo", ContractManager.dbPath);
            String tagA = (ps == System.out ? "[Contract_" + port + "_out] " : "");
            String tagB = (ps == System.out ? "[Contract_" + port + "_err] " : "");
            outputTracer.track(process, sc, tagA, ps);
            errorTracer.track(process, new Scanner(process.getErrorStream()), tagB, ps);
            get.syncGet("", "registerMangerPort", String.valueOf(ContractManager.cPort.getCMPort()));

            if (isBundlePath(contractMeta.contract.getScriptStr())) {
                status =
                        get.syncGet(
                                "", "setContractBundle", JsonUtil.toJson(contractMeta.contract));
            } else {
                status = get.syncGet("", "setContract", JsonUtil.toJson(contractMeta.contract));
            }
            LOGGER.debug("port:" + port + " status:" + status);
            ContractResult r = JsonUtil.fromJson(status, ContractResult.class);
            if (r.status == Status.Success) {
                initProps();
                get.syncGet("", "setPID", pid);
            } else if (r.status == null) {
                r.status = Status.Error;
                r.result = new JsonPrimitive(status);
                status = JsonUtil.toJson(r);
                contractMeta.name = get.syncGet("", "getContractName", "");
            }
            return status;
        } catch (Exception e) {
            e.printStackTrace();
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            e.printStackTrace(new PrintStream(bo));
            ContractResult r =
                    new ContractResult(Status.Exception, new JsonPrimitive(bo.toString()));
            return JsonUtil.toJson(r);
        }
    }

    public void updateMemory() {
        try {
            if (System.currentTimeMillis() - lastUpdate < 2000) {
                return;
            }
            lastUpdate = System.currentTimeMillis();
            get.asyncGet(
                    "",
                    ".UsedMemory",
                    "",
                    new ResultCallback() {
                        @Override
                        public void onResult(String str) {
                            memory = Long.parseLong(str);
                        }
                    });
        } catch (Exception e) {
            // e.printStackTrace();
        }
    }

    public Map<String, REventSemantics> getEvents() {
        return contractMeta.declaredEvents;
    }

    public Collection<FunctionDesp> getExportedFunctions() {
        return contractMeta.getExportedFunctions();
    }

    public String getIdentifier() {
        return get.syncGet("", "getIdentifier", " ");
    }

    public void setIdentifier(String alias) {
        get.asyncGet("", "setIdentifier", alias, null);
    }

    public boolean isUnit() {
        return contractMeta.contract.getType() != ContractExecType.Sole;
    }

    public String signResult(String result) {
        return contractMeta.contract.signResult(result);
    }

    public String getPubkey() {
        return contractMeta.contract.getPublicKey();
    }

    public String getContractID() {
        return contractMeta.contract.getID();
    }

    public String getContractName() {
        return contractMeta.name;
    }

    public String getContractDOI() {
        return contractMeta.contract.getDOI();
    }

    public String getContractKey() {
        return contractMeta.contract.getKey();
    }

    public void setContractKey(String key) {
        contractMeta.contract.setKey(key);
    }

    public ContractExecType getContractType() {
        return contractMeta.contract.getType();
    }

    public List<AnnotationNode> getAnnotations() {
        return contractMeta.annotations;
    }

    public int getContractCopies() {
        return contractMeta.contract.getNumOfCopies();
    }

    public String getLogType(String action) {
        return contractMeta.logDetail.get(action);
    }

    public boolean isDebug() {
        return contractMeta.isDebug;
    }

    public long getTimes() {
        return times;
    }

    public long getTraffic() {
        return traffic;
    }

    public YjsType getYjsType() {
        return contractMeta.contract.getYjsType();
    }

    public String getPID() {
        return pid;
    }

    public String getPort() {
        return "" + port;
    }

    public String getMemory() {
        return "" + memory;
    }

    public void loadTimesAndTraffic() {
        String tempTime2 =
                KeyValueDBUtil.instance.getValue(
                        CMTables.ContractInfo.toString(), contractMeta.name + "-Times");
        String tempTraffic2 =
                KeyValueDBUtil.instance.getValue(
                        CMTables.ContractInfo.toString(), contractMeta.name + "-Traffic");
        if (tempTime2 != null && !tempTime2.equals("")) {
            times = Long.parseLong(tempTime2);
        }
        if (tempTraffic2 != null && !tempTraffic2.equals("")) {
            traffic = Long.parseLong(tempTraffic2);
        }
    }

    public void saveTimesAndTraffic() {
        KeyValueDBUtil.instance.setValue(
                CMTables.ContractInfo.toString(), contractMeta.name + "-Times", times + "");
        KeyValueDBUtil.instance.setValue(
                CMTables.ContractInfo.toString(), contractMeta.name + "-Traffic", traffic + "");
    }

    public void setMask(JsonObject args) {

        //get.asyncGet("",,,,);
        get.asyncGet("", "setMask", args.toString(), null);

    }

    public void setProjectConfig(String args) {
        get.asyncGet("", "setProjectConfig", args, null);
    }
    //public String

    static class ReqScript {
        String mode;
        String code;
    }
}
