package org.bdware.sc;

import org.bdware.sc.bean.Contract;
import org.bdware.sc.bean.FunctionDesp;
import org.bdware.sc.bean.IDSerializable;
import org.bdware.sc.event.REvent;
import org.bdware.sc.node.AnnotationNode;
import org.bdware.sc.node.YjsType;

import java.util.*;

public class ContractMeta implements IDSerializable {
    public Contract contract;
    ContractStatusEnum status;
    String name;
    String id;
    boolean isDebug;
    Map<String, REvent.REventSemantics> declaredEvents;
    Map<String, FunctionDesp> exportedFunctions;
    Map<String, String> logDetail;
    List<AnnotationNode> annotations;
    Set<String> dependentContracts;
    transient Map<String, FunctionDesp> funCache;
    boolean sigRequired;
    String thisPermission; // 合约当前权限
    /*
    {
       "name": "dx_substr",
       "parameter":
       {
           "columnIndex":5,
           "paras":["1","3"]
       }
    },
    */
    // Map<Object,Object>MaskInfo;

    public ContractMeta() {
    }

    public ContractStatusEnum getStatus() {
        return status;
    }

    public void setStatus(ContractStatusEnum status) {
        this.status = status;
    }

    public String getID() {
        return id;
    }

    public List<AnnotationNode> getAnnotations() {
        return annotations;
    }

    public Collection<FunctionDesp> getExportedFunctions() {
        return null == exportedFunctions ? null : exportedFunctions.values();
    }

    public Set<String> getDependentContracts() {
        return dependentContracts;
    }

    public Map<String, REvent.REventSemantics> getEvents() {
        return declaredEvents;
    }

    public YjsType getYjsType() {
        return contract.getYjsType();
    }

    public String getPubkey() {
        return contract.getPublicKey();
    }

    public String getThisPermission() {
        return thisPermission;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    // public setMask(){}

    public boolean getIsDebug() {
        return isDebug;
    }

    public FunctionDesp getExportedFunction(String action) {
        if (funCache == null) {
            funCache = new HashMap<>();
        }
        FunctionDesp desp = funCache.get(action);
        if (desp != null) {
            return desp;
        }
        desp = seekFunction(action);
        if (null != desp) {
            funCache.put(action, desp);
        }
        return desp;
    }

    public FunctionDesp seekFunction(String action) {
        return null == exportedFunctions ? null : exportedFunctions.get(action);
    }

    public void setContract(Contract c) {
        contract = c;
        id = c.getID();
        status = ContractStatusEnum.HANGED;
        isDebug = false;
    }
}
