package org.bdware.sc.event;

import org.bdware.sc.util.JsonUtil;

import static org.bdware.sc.ContractManager.instance;

/**
 * @author Kaidong Wu
 */
public class EventCenter {
//    private static final Logger LOGGER = LogManager.getLogger(EventCenter.class);

    /**
     * get the nearest node to the topic in the hash function range
     *
     * @param topic the topic
     * @return id of the node
     */
    public String getCenterByTopic(String topic) {
        if (null == instance.nodeCenterConn) {
            return null;
        }
        String[] centers = instance.nodeCenterConn.getClusterByKey(topic, 1);
        if (null == centers) {
            return null;
        }
        return centers[0];
    }

    /**
     * subscribe a topic in center
     *
     * @param topic     event topic
     * @param semantics event semantics, used to mark PRESUB events
     */
    public void subInCenter(String topic, REvent.REventSemantics semantics) {
        if (null == instance.nodeCenterConn) {
            return;
        }
        REvent msg = new REvent(
                topic,
                REvent.REventType.SUBSCRIBE,
                String.format("{\"subscriber\":\"%s\"}",
                        instance.nodeCenterConn.getNodeId(null)),
                "");
        msg.setSemantics(semantics);
        msg.doSignature(instance.nodeCenterConn.getNodeKeyPair());
        String nodeId = getCenterByTopic(topic);
        instance.nodeCenterConn.deliverEvent(JsonUtil.toJson(msg), nodeId);
    }

    /**
     * deliver an event to center
     *
     * @param topic event topic
     * @param event the event
     * @return if this node is the center, return false; otherwise, return true
     */
    public boolean deliverEvent(String topic, REvent event) {
        if (null == instance.nodeCenterConn) {
            return false;
        }
        String nodeId = getCenterByTopic(topic);
        return instance.nodeCenterConn.deliverEvent(JsonUtil.toJson(event), nodeId);
    }

    /**
     * publish an event to another node; used by NodeConsumer
     *
     * @param eStr   event string
     * @param target id of the target node
     */
    public void publishEvent(String eStr, String target) {
        if (null != instance.nodeCenterConn) {
            instance.nodeCenterConn.deliverEvent(eStr, target);
        }
    }
}