package org.bdware.sc.handler;

import org.bdware.sc.ContractManager;
import org.bdware.sc.bean.Contract;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.Description;
import org.bdware.sc.conn.MsgHandler;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.event.REvent;
import org.bdware.sc.get.GetMessage;
import org.bdware.sc.util.JsonUtil;

public class ManagerHandler extends MsgHandler {

    private final ContractManager cm;

    public ManagerHandler(ContractManager contractManager) {
        cm = contractManager;
    }

    @Description(value = "start Contract, {\"type\":\"Data\",\"id\":\"123bbaa\"}", isAsync = true)
    public void startContract(GetMessage msg, ResultCallback cb) {
        Contract c = JsonUtil.fromJson(msg.arg, Contract.class);
        cb.onResult(cm.startContract(c));
    }

    @Description("request Contract, TODO")
    public void requestContract(GetMessage msg, ResultCallback cb) {
        ContractRequest c = JsonUtil.fromJson(msg.arg, ContractRequest.class);
        cb.onResult(cm.requestContract(c));
    }

    @Description("get how many times a contract has been executed")
    public void getTimesOfExecution(GetMessage msg, ResultCallback cb) {
        cb.onResult(cm.getTimesOfExecution(msg.arg));
    }

    @Description(
            value = "execute Contract, {\"contractID\":\"112233\",\"arg\":\"\"}",
            isAsync = true)
    public void executeContract(GetMessage msg, ResultCallback cb) {
        ContractRequest c;
        try {
            c = JsonUtil.fromJson(msg.arg, ContractRequest.class);
            cm.executeContractInternal(c, cb, null);
        } catch (Exception ignored) {
        }
    }


    @Description("stop Contract, {\"id\":\"112233\"}")
    public void stopContract(GetMessage msg, ResultCallback cb) {
        ContractRequest c = JsonUtil.fromJson(msg.arg, ContractRequest.class);
        cb.onResult(cm.stopContract(c.getContractID()));
    }

    @Description("list Contracts")
    public void listContracts(GetMessage msg, ResultCallback cb) {
        cb.onResult(cm.listContracts(null));
    }

    @Description("stop all Contracts")
    public void stopAllContracts(GetMessage msg, ResultCallback cb) {
        int count = cm.stopAllContracts();
        cb.onResult(count + " contracts stoped!");
    }

    @Description("exit ContractManager!")
    public void exit(GetMessage msg, ResultCallback cb) {
        ContractManager.threadPool.execute(() -> {
            System.out.println("ManagerHandler: exit in 3 seconds!");
            try {
                Thread.sleep(3000);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            System.exit(0);
        });
        cb.onResult("success");
    }

    @Description("query distributed execution port by contractID/name")
    public void queryDEPort(GetMessage msg, ResultCallback cb) {
        cb.onResult(cm.queryDEPort(msg.arg));
    }

    @Description("add distributed execution member, sample: {contractID:id/name, arg:ipAndPort}")
    public void addDEMember(GetMessage msg, ResultCallback cb) {
        ContractRequest cr = JsonUtil.fromJson(msg.arg, ContractRequest.class);
        cb.onResult(cm.addDEMember(cr.getContractID(), cr.getArg()));
    }

    /**
     * @author Kaidong Wu
     */
    @Description("Deliver event message")
    public void deliverEMessage(GetMessage msg, ResultCallback cb) {
        REvent eMsg = JsonUtil.fromJson(msg.arg, REvent.class);
        cb.onResult(cm.deliverEMessage(eMsg));
    }
}
